/*****************************************************************************
THIS FILE CONTAINS THE C FUNCTION CALLED IN R

   - runs DSA on each training set
   - uses empirical RSS to form models
   - uses cross-validation to pick size and order of interaction of final model
   - runs swap-sub. moves if complexity measure is exceeded
   - incorporates rules on what variables cannot interact
   - fits weighted regression models
   - automatically fills cross-validated risks grid for maxsize and orderint

  
--Counters and Constants
gCOUNT:
0 number of terms in the last model accepted - intercept excluded (msize)
1 current maximum order of interaction allowed (orderint)
2 number of calls to glm routine (nglmcall)
3 fold number (fold)
4 size of training set (ntrain)
5 size of validation set (nval)

gCONST:
0 max number of terms allowed - intercept excluded (maxsize)
1 max interaction order (maxorderint)
2 max sum of powers in each term (maxsumofpow)
3 number of data splits (nfolds)
4 total number of observations in learning set (nlearn)
5 number of candidate variables (nvarX)
6 indicator of weight use (CENSOR)
7 optimal size chosen via cross-validation (bestsize)
8 optimal order of interaction chosen via cross-validation (bestorderint)
9 indicator of the outcome being binary (binind)
***************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include <time.h>
#include <R.h>
#include <R_ext/Utils.h>
#include "dsa.h"

void DSA_PACK_Rentry(int *gCONST,double *Xlearn,double *Ylearn,double *CVrisks,double *PARA,
		     int *vsplits,double *WTtrainvallearn,int *nforced,int *forcedterms,
		     int *maxntrain, int *maxnval,int *modelselected)
{
  int i,j,k,maxsize,orderint,maxsumofpow,maxorderint,bestsize,bestorderint,nfolds,fold,totalglmcalls,nlearn,nvarX,CENSOR,**bestmodels,ntrain,nval,binind,*trainindex,*valindex,*sx,*bestmodelsspace,*currentmodel,*tempmodel,*nextmodel,*newterm,*ithterm,*gCOUNT; 

  double grandmin,meanCVrisk,*CVrisksbysplit,*SE,*COV,*WK,*glmWT,*H,*RES,*Q,*PP,*bestarss,*currentarss,*Ytrain,*Xtrain,*Yval,*Xval,*tempXval,*tempYval,*temptermval,*temptermlearntrain,*tempXlearntrain,*tempYlearntrain,*WTlearntrain,*WTval;

  /*  Rprintf("\nDSA package version 1.3.2\n"); */

  /* INITIALIZATION */
  maxsize = gCONST[0] ;        /* maxsize: max number of terms allowed - intercept excluded */
  maxorderint = gCONST[1] ;    /* maxorderint: max interaction order */
  maxsumofpow = gCONST[2] ;    /* maxsumofpow: max sum of powers in each term */
  nfolds = gCONST[3] ;         /* nfolds: number of data splits */
  nlearn = gCONST[4] ;         /* nlearn: total number of observations in learning set */
  nvarX = gCONST[5] ;          /* nvarX: number of candidate variables*/
  CENSOR = gCONST[6] ;         /* CENSOR: indicator of weight use */
                               /* gCONST[7] = bestsize: optimal size chosen via cross-validation */
                               /* gCONST[8] = bestorderint: optimal order of interaction chosen via cross-validation */
  binind = gCONST[9] ;         /* binind: indicator of the outcome being binary */
  totalglmcalls = 0;


  /* MEMORY ALLOCATION */ /* NEED TO MAKE SURE TO INITIALIZE IF NECESSARY */
  trainindex=(int *)R_alloc(*maxntrain,sizeof(int));
  valindex=(int *)R_alloc(*maxnval,sizeof(int));
  CVrisksbysplit=(double *)R_alloc((maxsize+1)*nfolds,sizeof(double));
  SE=(double *)R_alloc(maxsize+1,sizeof(double));
  COV=(double *)R_alloc((maxsize+1)*(maxsize+2)/2,sizeof(double));
  glmWT=(double *)R_alloc(nlearn,sizeof(double));
  RES=(double *)R_alloc(nlearn,sizeof(double));

  double* new_coefficients;
  double* WX;
  double* Z;
  double* qraux;
  double* effects;
  double* work;
  int* pivot;
  int pivot_i;

  if (binind == 1) {
    new_coefficients = (double *) R_alloc(maxsize+1, sizeof(double)); 
    WX = (double *) R_alloc((maxsize + 1) * nlearn, sizeof(double)); 
    Z = (double *) R_alloc(nlearn, sizeof(double)); 
  }
  qraux = (double *) R_alloc(maxsize+1, sizeof(double)); 
  effects = (double *) R_alloc(nlearn, sizeof(double));
  pivot = (int *) R_alloc(maxsize+1, sizeof(int));
  work  = (double *) R_alloc(2*(maxsize+1), sizeof(double));
    
  for (pivot_i = 0; pivot_i < (maxsize+1); pivot_i++) {
    pivot[pivot_i] = pivot_i + 1;
  }
  
  sx=(int *)R_alloc(maxsize+1,sizeof(int));
  bestarss=(double *)R_alloc(maxsize+1,sizeof(double));
  currentarss=(double *)R_alloc(maxsize+1,sizeof(double));
  Ytrain=(double *)R_alloc(*maxntrain,sizeof(double));

  Xtrain=(double *)R_alloc(*maxntrain*nvarX,sizeof(double));
  memset(Xtrain, 0, sizeof(double)*(*maxntrain)*nvarX); /** JHB valgrind. **/

  Yval=(double *)R_alloc(*maxnval,sizeof(double));

  Xval=(double *)R_alloc(*maxnval*nvarX,sizeof(double));
  memset(Xval, 0, sizeof(double)*(*maxnval) * nvarX); /** JHB valgrind. **/

  tempXval=(double *)R_alloc(*maxnval*(maxsize+1),sizeof(double)); 
  memset(tempXval, 0, sizeof(double)*(*maxnval) * (maxsize+1)); /** JHB valgrind. **/

  tempYval=(double *)R_alloc(*maxnval,sizeof(double));
  temptermval=(double *)R_alloc(*maxnval,sizeof(double));
  temptermlearntrain=(double *)R_alloc(nlearn,sizeof(double));

  tempXlearntrain=(double *)R_alloc(nlearn*(maxsize+1),sizeof(double));
  memset(tempXlearntrain, 0.0, sizeof(double)*nlearn*(maxsize+1)); /** JHB valgrind. **/

  tempYlearntrain=(double *)R_alloc(nlearn,sizeof(double));
  bestmodelsspace=(int *)R_alloc(maxsize*maxsize*nvarX,sizeof(int));
  bestmodels=(int **)R_alloc(maxsize,sizeof(int *));
  for(i=0;i<maxsize;i++)*(bestmodels+i) = bestmodelsspace+i*(maxsize*nvarX);
  currentmodel=(int *)R_alloc(maxsize*nvarX,sizeof(int));
  tempmodel=(int *)R_alloc(maxsize*nvarX,sizeof(int));
  nextmodel=(int *)R_alloc(maxsize*nvarX,sizeof(int));
  newterm=(int *)R_alloc(nvarX,sizeof(int));
  ithterm=(int *)R_alloc(nvarX,sizeof(int));
  WTlearntrain=(double *)R_alloc(nlearn,sizeof(double));
  WTval=(double *)R_alloc(*maxnval,sizeof(double));
  gCOUNT=(int *)R_alloc(6,sizeof(int));

  /* START DATA-ADAPTIVE SELECTION OF BEST SIZE AND ORDER OF INTERACTION */
  for(orderint = 1; orderint <= maxorderint; orderint++)
    {
      for(fold = 0; fold < nfolds; fold++)
	{
	  /* COMPUTE FOR THE CORRESPONDING SPLIT THE SIZE OF THE TRAINING AND VALIDATION SET */
	  nval=0;
	  for(i=0;i<nlearn;i++)if(*(vsplits+fold+i*nfolds)==1)nval++;
	  ntrain=nlearn-nval;
	  if(DSA_PACK_userMlevel >= DSA_PACK_Mbase)
	    Rprintf("\n****** Model selection among models of maximum interaction order = %li based on split V%li ******",
		    orderint,fold+1); 

	  /*IDENTIFIES WHICH OBSERVATION IS IN THE TRAIN AND VAL SETS FOR THE CURRENT SPLIT OF THE DATA*/
	  DSA_PACK_initialize_l(trainindex,ntrain);
	  DSA_PACK_initialize_l(valindex,nval);
	  j=0;
	  k=0;
	  for(i=0; i<nlearn; i++)
	    {
	      if(vsplits[i*nfolds+fold] == 0)
		{
		  trainindex[j] = i;
		  j++;
		}
	      else
		{
		  valindex[k] = i;
		  k++;
		}
	    }

   	  /* CREATE TWO MATRICES CONTAINING THE TRAINING DATA: Y AND X */
	  for(i = 0; i < ntrain; i++)
	    {
	      Ytrain[i] = Ylearn[trainindex[i]];
	      WTlearntrain[i] = WTtrainvallearn[trainindex[i]*(nfolds+1) + fold];
	      for(j = 0; j < nvarX; j++)Xtrain[i*nvarX+j] = Xlearn[trainindex[i]*nvarX + j];
	    }
	  /* CREATE TWO MATRICES CONTAINING THE VALIDATION DATA: Y AND X */
	  for(i = 0; i < nval; i++)
	    {
	      Yval[i] = Ylearn[valindex[i]];
	      WTval[i] = WTtrainvallearn[valindex[i]*(nfolds+1) + fold];
	      for(j=0; j<nvarX; j++)Xval[i*nvarX+j] = Xlearn[valindex[i]*nvarX + j];
	    }

	  gCOUNT[0] = 0;              /* msize: number of terms in the current model - intercept excluded */
	  gCOUNT[1] = orderint;       /* orderint: current maximum order of interaction allowed */
	  gCOUNT[2] = 0;              /* nglmcall: number of calls to glm routine */
	  gCOUNT[3] = fold;           /* fold: fold number (fold) */ 
	  gCOUNT[4] = ntrain;         /* ntrain: size of training set */
	  gCOUNT[5] = nval;           /* nval: size of validation set */

	  if(DSA_PACK_userMlevel>=DSA_PACK_Mlow)
	    {
	      Rprintf("\n");
	      Rprintf("\tMaximal number of terms in each model considered - intercept excluded = %ld \n",
		      maxsize);
	      Rprintf("\tMaximal sum of powers in each term of each model considered = %ld \n",maxsumofpow);
	      Rprintf("\tSize of the validation set (nval): %ld \n",nval);
	      Rprintf("\tSize of the training set (ntrain): %ld \n",ntrain);
	      Rprintf("\n");
	    }

	  /* RUN THE DSA ALGORITHM */
	  if(DSA_PACK_userMlevel>=DSA_PACK_Mbase)
	    Rprintf("Selecting models on the training set based on average risks.\n");

	  DSA_PACK_startAlgo(glmWT,PARA,SE,COV,RES,gCOUNT,gCONST,bestarss,Ytrain,
			     Xtrain,sx,bestmodels,currentmodel,tempmodel,nextmodel,newterm,ithterm,
			     temptermlearntrain,tempXlearntrain,tempYlearntrain,WTlearntrain,1,currentarss,
			     nforced,forcedterms,
			     new_coefficients, WX,  Z,  qraux, effects, work, pivot);

	  if(DSA_PACK_userMlevel>=DSA_PACK_Mlow)
	    Rprintf("The largest model considered is of size %li.\n",gCOUNT[0]);
	  
	  totalglmcalls = totalglmcalls + gCOUNT[2];
	  if(DSA_PACK_userMlevel>=DSA_PACK_Mlow)
	    Rprintf("The number of times glm was called on this training set is %ld.\n",gCOUNT[2]);

	  /* COMPUTE THE CROSS-VALIDATED RISKS */
	  if(DSA_PACK_userMlevel>=DSA_PACK_Mlow)
	    Rprintf("Computing the average cross-validated risks of the models selected.\n");
	  gCOUNT[2] = 0; /* reset nglmcall */
	  
	  DSA_PACK_getaverageCVrisks(gCOUNT,gCONST,sx,Xtrain,Ytrain,Xval,Yval,bestmodels,tempXlearntrain,
				     tempYlearntrain,tempXval,tempYval,temptermlearntrain,temptermval,
				     PARA,SE,COV, RES,glmWT,CVrisksbysplit,WTlearntrain,WTval,*nforced,
				     tempmodel, new_coefficients, WX,  Z,  qraux, effects, work, pivot);

				     
	  totalglmcalls = totalglmcalls + gCOUNT[2];
	  if(DSA_PACK_userMlevel>=DSA_PACK_Mlow)
	    Rprintf("The number of times glm was called on this validation set is %ld.\n",gCOUNT[2]);
	}

      if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)
	Rprintf("The average cross-validated risks by split (maxsize by nfolds) are: \n");

      if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)
	DSA_PACK_printmatrix(CVrisksbysplit,maxsize+1,nfolds);

      /* AVERAGE CV RISKS ACROSS VALIDATION SETS */
      for(i=0;i<maxsize+1;i++)
	{     
	  meanCVrisk=0;
	  for(j=0;j<nfolds;j++)meanCVrisk += CVrisksbysplit[j*(maxsize+1)+i];
	  meanCVrisk/=nfolds;
	  CVrisks[i*maxorderint+(orderint-1)] = meanCVrisk;
	}
    }
  if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)
    Rprintf("\nThe cross-validated risks averaged over each split (maxorderint by maxsize) are: \n");
  if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)DSA_PACK_printmatrix(CVrisks,maxorderint,maxsize+1);

  /* PICK BEST CV RISK TO PICK SELECT BEST SIZE AND ORDER OF INTERACTION */
  grandmin = DSA_PACK_getmin(CVrisks,maxorderint*(maxsize+1));
  bestsize=bestorderint=-1;
  for(i=0;i<maxorderint;i++)
    for(j=0;j<maxsize+1;j++)
      if(CVrisks[j*maxorderint+i]!=R_PosInf)
	if(abs(CVrisks[j*maxorderint+i] - grandmin)<epsiloncompare)
	  {
	    bestsize = j ;
	    bestorderint = i + 1;
	    i=maxorderint+1;
	    break;
	  }
  if(DSA_PACK_userMlevel>=DSA_PACK_Mlow)
    Rprintf("\nMinimum average cross-validated risk of %lf is reached for a model of size %li with a maximum order of interaction of %li.\n",grandmin,bestsize,bestorderint);
  gCONST[7]=bestsize;
  gCONST[8]=bestorderint;

  /* GET FINAL BEST MODEL BY RUNNING DSA ON LEARNING SET */
  if(DSA_PACK_userMlevel>=DSA_PACK_Mbase)
    Rprintf("\n****** Selecting the best model of size = %li and maximum order of interaction = %li on the learning set based on average residuals sum of squares ******\n",gCONST[7],gCONST[8]);
  gCOUNT[2] = 0;   /* reset nglmcall */  
  for(i = 0; i < nlearn; i++)
    WTlearntrain[i] = WTtrainvallearn[i*(nfolds+1) + nfolds];
  
  DSA_PACK_startAlgo(glmWT,PARA,SE,COV,RES,gCOUNT,gCONST,bestarss,Ylearn,Xlearn,sx,
		     bestmodels,currentmodel,tempmodel,nextmodel,newterm,ithterm,temptermlearntrain,
		     tempXlearntrain,tempYlearntrain,WTlearntrain,0,currentarss,nforced,forcedterms,
		     new_coefficients, WX,  Z,  qraux, effects, work, pivot);
  

  totalglmcalls = totalglmcalls + gCOUNT[2];

  if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)
    Rprintf("\nThe number of times glm was called on the learning set is %ld.\n",gCOUNT[2]);
  if(DSA_PACK_userMlevel>=DSA_PACK_Mmedium)
    Rprintf("The total number of calls to glm is %ld.\n",totalglmcalls);


  /* OUTPUT THE MODEL SELECTED */
  for(i=0;i<maxsize;i++)
    for(j=0;j<nvarX;j++)
      *(modelselected+i+j*maxsize)=*(*(bestmodels+bestsize-1)+i+j*maxsize);
  
  if(DSA_PACK_userMlevel>=DSA_PACK_Mlow)Rprintf("\nLeaving the C subroutine.\n");
}

void DSA_PACK_setMessagelevel(int *newMlevel)
{
  int oldlevel=DSA_PACK_userMlevel;

  DSA_PACK_userMlevel=*newMlevel;
}
